/* This file is part of the wvWare 2 project
   Copyright (C) 2001 Werner Trobin <trobin@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "olestorage.h"
#include "olestream.h"
#include <iostream>

using namespace wvWare;

OLEStorage::OLEStorage() : m_storage( 0 ), m_lastError( MS_OLE_ERR_OK ),
    m_fileName( "" ), m_path( "/" )
{
}

OLEStorage::OLEStorage( const std::string& fileName ) : m_storage( 0 ),
    m_lastError( MS_OLE_ERR_OK ), m_fileName( fileName ), m_path( "/" )
{
}

OLEStorage::~OLEStorage()
{
    close();  // just in case
}

void OLEStorage::setName( const std::string& fileName )
{
    if ( m_storage )  // don't set a new name if we already have a storage
        return;
    m_fileName = fileName;
}

bool OLEStorage::open()
{
    // We already have an open storage, return
    if ( m_storage )
        return true;

    if ( m_fileName.empty() )
        return false;

    // try to open the file (r+w or at least r)
    m_lastError = ms_ole_open_vfs( &m_storage, m_fileName.c_str(), true, 0 );

    if ( m_lastError == MS_OLE_ERR_BADARG ||  // impossible?
         m_lastError == MS_OLE_ERR_FORMAT )   // not an OLE storage
        return false;
    else if ( m_lastError == MS_OLE_ERR_EXIST ) { // file doesn't exist -> try to create it
        m_lastError = ms_ole_create_vfs( &m_storage, m_fileName.c_str(), true, 0 );
        if ( m_lastError == MS_OLE_ERR_BADARG ||  // impossible?
             m_lastError == MS_OLE_ERR_PERM ||    // no permissions
             m_lastError == MS_OLE_ERR_SPACE )    // not enough space
            return false;
    }
    return true;
}

void OLEStorage::close()
{
    if ( !m_storage )  // already closed?
        return;

    // check if we still have some open streams and close them properly
    std::list<OLEStream*>::const_iterator it = m_streams.begin();
    std::list<OLEStream*>::const_iterator end = m_streams.end();
    while ( it != end ) {
        OLEStream* stream = *it;
        ++it; // first advance the iterator, as the stream will remove itself from the list
        delete stream;
        std::cerr << "Warning: Closing a stream you didn't delete." << std::endl;
    }
    m_streams.clear();
    ms_ole_destroy( &m_storage );  // finally close it
    m_storage = 0; // trust no one :p
}

std::list<std::string> OLEStorage::listDirectory()
{
    char** nodes;
    std::list<std::string> ret;

    m_lastError = ms_ole_directory( &nodes, m_storage, m_path.c_str() );

    if ( m_lastError == MS_OLE_ERR_BADARG || // huh?
         m_lastError == MS_OLE_ERR_INVALID )  // not an ole file??
        return ret;

    int i = 0;
    char* node = nodes[i];
    while ( node ) {
        ret.push_back( node );
        g_free( node );  // we got ownership :}
        node = nodes[ ++i ];
    }
    g_free( nodes ); // we got ownership :}
    return ret;
}

bool OLEStorage::enterDirectory( const std::string& directory )
{
    MsOleStat s = stat( directory.c_str() );
    if ( s.type != MsOleStorageT || m_lastError != MS_OLE_ERR_OK )
        return false;
    m_path += directory + "/";
    return true;
}

void OLEStorage::leaveDirectory()
{
    if ( !m_storage )
        return;
    // we never want to get rid of the trailing '/'
    m_path.resize( m_path.find_last_of( '/', m_path.length() - 2 ) + 1 );
    return;
}

bool OLEStorage::setPath( const std::string& path )
{
    std::string oldPath = m_path, tmp;
    m_path = "/";
    std::string::size_type start = 1, end = 0;
    bool success = true;

    if ( path[ 0 ] != '/' )
        start = 0;

    while ( start < path.length() && success ) {
        end = path.find_first_of( '/', start );
        if ( end != std::string::npos ) {
            tmp = path.substr( start, end - start );
            start = end + 1;
        }
        else {
            tmp = path.substr( start );
            start = std::string::npos;
        }
        if ( !enterDirectory( tmp ) )
            success = false;
    }

    if ( !success )
        m_path = oldPath;  // restore the old path
    return success;
}

MsOleStat OLEStorage::stat( const std::string& node )
{
    MsOleStat result;
    result.type = MsOleRootT;
    result.size = static_cast<unsigned int>( -1 );

    if ( !m_storage )
        return result;

    m_lastError = ms_ole_stat( &result, m_storage, m_path.c_str(), node.c_str() );
    return result;
}

OLEStreamReader* OLEStorage::createStreamReader( const std::string& stream )
{
    OLEStreamReader* reader = 0;
    MsOleStream* s = 0;

    // Open the stream for reading
    m_lastError = ms_ole_stream_open( &s, m_storage, m_path.c_str(), stream.c_str(), 'r' );
    if ( s && m_lastError == MS_OLE_ERR_OK ) {
        reader = new OLEStreamReader( s, this );
        m_streams.push_back( reader );
    }
    return reader;
}

OLEStreamWriter* OLEStorage::createStreamWriter( const std::string& stream )
{
    OLEStreamWriter* writer = 0;
    MsOleStream* s = 0;

    // Don't try to confuse our highly intelligent path system :p
    if ( stream.find('/') != std::string::npos ) {
        std::cerr << "Warning: You tried to create a stream with a '/' in its name." << std::endl;
        return 0;
    }

    // Open the stream for writing
    m_lastError = ms_ole_stream_open( &s, m_storage, m_path.c_str(), stream.c_str(), 'w' );
    if ( s && m_lastError == MS_OLE_ERR_OK ) {
        writer = new OLEStreamWriter( s, this );
        m_streams.push_back( writer );
    }
    return writer;
}

void OLEStorage::debug() const
{
    // 0 ... most verbose option
    // 1 ... semi-verbose
    // 2 ... just the tree
    if ( m_storage )
        ms_ole_debug( m_storage, 0 );
}

void OLEStorage::streamDestroyed( OLEStream* stream )
{
    m_streams.remove( stream );
}
