/* This file is part of the wvWare 2 project
   Copyright (C) 2001 Werner Trobin <trobin@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "parser95.h"
#include "olestream.h"
#include "word97_helper.h"
#include "styles.h"
#include "textconverter.h"
#include "convert.h"
#include "handlers.h"
#include "fonts.h"
#include "lists.h"
#include "associatedstrings.h"
#include <iostream>

using namespace wvWare;

Parser95::Parser95( OLEStorage* storage, OLEStreamReader* wordDocument ) :
    Parser( storage, wordDocument ), Parser9x(), m_data( 0 ), m_stylesheet( 0 ),
    m_fonts( 0 ), m_lists( 0 ), m_fib( wordDocument, true ),
    m_fib97( Word95::toWord97( m_fib ) ), // make the Word97 struct out of the Word95 one
    m_plcfpcd( 0 ), m_plcfsed( 0 ), m_plcfbtePapx( 0 ), m_plcfbteChpx( 0 ),
    m_currentCP( 0 ), m_textconverter( 0 )
{
    if ( !isOk() )
        return;

    m_data = storage->createStreamReader( "Data" );
    if ( !m_data || !m_data->isValid() ) {
        std::cerr << "Warning: Couldn't open the Data stream" << std::endl;
        delete m_data;
        m_data = 0;
    }

#ifdef WV2_DUMP_FIB
    std::cerr << "Dumping some parts of the FIB: " << std::endl;
    std::cerr << "   wIdent=" << m_fib.wIdent << std::endl;
    std::cerr << "   nFib=" << m_fib.nFib << std::endl;
    std::cerr << "   nFibBack=" << m_fib.nFibBack << std::endl;
    std::cerr << "   lid=0x" << std::hex << m_fib.lid << std::dec << std::endl;
//    std::cerr << "   lidFE=0x" << std::hex << m_fib.lidFE << std::dec << std::endl;
    std::cerr << "   fEncrypted=" << m_fib.fEncrypted << std::endl;
//    std::cerr << "   chs=" << m_fib.chs << std::endl;
    std::cerr << "   fcMin=" << m_fib.fcMin << std::endl;
    std::cerr << "   fcMac=" << m_fib.fcMac << std::endl;
    std::cerr << "   ccpText=" << m_fib.ccpText << std::endl;
    std::cerr << "   ccpFtn=" << m_fib.ccpFtn << std::endl;
    std::cerr << "   ccpHdd=" << m_fib.ccpHdd << std::endl;
    std::cerr << "   ccpMcr=" << m_fib.ccpMcr << std::endl;
    std::cerr << "   ccpAtn=" << m_fib.ccpAtn << std::endl;
    std::cerr << "   ccpEdn=" << m_fib.ccpEdn << std::endl;
    std::cerr << "   ccpTxbx=" << m_fib.ccpTxbx << std::endl;
    std::cerr << "   ccpHdrTxbx=" << m_fib.ccpHdrTxbx << std::endl;
//    std::cerr << "   pnFbpChpFirst=" << m_fib.pnFbpChpFirst << std::endl;
    std::cerr << "   pnChpFirst=" << m_fib.pnChpFirst << std::endl;
    std::cerr << "   cpnBteChp=" << m_fib.cpnBteChp << std::endl;
//    std::cerr << "   pnFbpPapFirst=" << m_fib.pnFbpPapFirst << std::endl;
    std::cerr << "   pnPapFirst=" << m_fib.pnPapFirst << std::endl;
    std::cerr << "   cpnBtePap=" << m_fib.cpnBtePap << std::endl;
#endif

    init();
}

Parser95::~Parser95()
{
    delete m_lists;
    delete m_fonts;
    delete m_stylesheet;
    delete m_textconverter;
    delete m_plcfpcd;
    delete m_plcfsed;
    delete m_plcfbtePapx;
    delete m_plcfbteChpx;
    delete m_data;
}

bool Parser95::parse()
{
    if ( !isOk() )
        return false;

    if ( m_fib.fEncrypted ) {
        std::cerr << "Error: The document is encrypted." << std::endl;
        return false;
    }

    // Read the DOP
    S32 old = m_wordDocument->seek( m_fib.fcDop );
    m_dop.read( m_wordDocument, false );
    if ( m_wordDocument->tell() != old + static_cast<S32>( m_fib.lcbDop ) )
        std::cerr << "Warning: DOP has a different size than expected." << std::endl;

    // Make the Word97 struct out of the Word95 one, since that's what the Parser API wants
    m_dop97 = Word95::toWord97( m_dop );

    // Read the PLCF SED
    m_wordDocument->seek( m_fib.fcPlcfsed );
    m_plcfsed = new Word95::PLCF<Word95::SED, 12>( m_fib.lcbPlcfsed, m_wordDocument );

    // Note that we test for the presence of complex structure, rather than
    // m_fib.fComplex. This allows us to treat newer files which always seem
    // to have piece tables in a consistent manner.
    //
    // There is also the implication that without the complex structures, the
    // text cannot be in unicode form.
    if ( m_fib.lcbClx != 0 ) {
        // Read the PAPX and CHPX BTE PLCFs (to locate the appropriate FKPs)
        m_wordDocument->seek( m_fib.fcPlcfbtePapx );
        m_plcfbtePapx = new Word95::PLCF<Word95::BTE, 4>( m_fib.lcbPlcfbtePapx, m_wordDocument );
        m_wordDocument->seek( m_fib.fcPlcfbteChpx );
        m_plcfbteChpx=new Word95::PLCF<Word95::BTE, 4>( m_fib.lcbPlcfbteChpx, m_wordDocument );

        // Now get the piece table
        if ( !readPieceTable() )
            return false;

        // ... other stuff to do before we start parsing... ### TODO?

        // start parsing the body
        if ( !parseBody() )
            return false;
    }
    else {
        std::cerr << "Warning: Found a Word95 file without piece table!" << std::endl;
        return false;
    }
    return true;
}

const Word97::FFN& Parser95::font( S16 ftc ) const
{
    return m_fonts->font( ftc );
}

AssociatedStrings Parser95::associatedStrings()
{
    return AssociatedStrings( m_fib.fcSttbfAssoc, m_fib.lcbSttbfAssoc, m_fib.lid, m_wordDocument );
}

const StyleSheet& Parser95::styleSheet() const
{
    return *m_stylesheet;
}

void Parser95::init()
{
/*    if ( m_fib.fFarEast )
        m_textconverter = new TextConverter( m_fib.lidFE );
    else*/
    m_textconverter = new TextConverter( m_fib.lid );

    // First create the whole stylesheet information.
    m_stylesheet = new StyleSheet( m_wordDocument, m_fib.fcStshf, m_fib.lcbStshf );
    m_fonts = new FontCollection( m_wordDocument, m_fib97 );
    m_lists = new ListInfoProvider( m_stylesheet );
    // Stage 2 of stylesheet construction, now that we know the list information
    // (chicken-egg problem)
    m_stylesheet->createListInfo( *m_lists );
}

bool Parser95::readPieceTable()
{
    m_wordDocument->seek( m_fib.fcClx );
    // first skip the leading grpprl blocks, we'll re-read them
    // if we need them later (no caching here)
    U8 blockType = m_wordDocument->readU8();
    while ( blockType == Word95::clxtGrpprl ) {
        U16 size = m_wordDocument->readU16();
#ifdef WV2_DUMP_PIECE_TABLE
        std::cerr << "Found a clxtGrpprl (size=" << size << ")" << std::endl;
#endif
        m_wordDocument->seek( size, SEEK_CUR );
        blockType = m_wordDocument->readU8();
    }
    if ( blockType == Word95::clxtPlcfpcd ) {
        U32 size = m_wordDocument->readU32();
#ifdef WV2_DUMP_PIECE_TABLE
        std::cerr << "Found the clxtPlcfpcd (size=" << size << ")" << std::endl;
#endif
        m_plcfpcd = new Word95::PLCF<Word95::PCD, 8>( size, m_wordDocument, false );

#ifdef WV2_DUMP_PIECE_TABLE
        Word95::PLCFIterator<Word95::PCD, 8> it( *m_plcfpcd );
        for ( int i = 0; it.current(); ++it, ++i ) {
            std::cerr << "Piece Table Entry(" << i << "): " << std::endl;
            std::cerr << "   start: " << it.currentStart() << std::endl;
            std::cerr << "   lim: " << it.currentLim() << std::endl;
            U32 fc = it.current()->fc;
            U32 limit = it.currentRun() << 1;
            std::cerr << "   value: " << fc << std::endl;
            if ( fc & 0x40000000 ) {
                fc = ( fc & 0xbfffffff ) >> 1;
                limit >>= 1;
                std::cerr << "   value (cleared 2nd MSB, div. by 2): " << fc << std::endl;
            }
            m_wordDocument->seek(fc);
            std::cerr << "   position: " << m_wordDocument->tell() << ", limit: " << limit << std::endl;
            for ( unsigned int j = 0; j < limit; ++j ) {
                U8 foo = m_wordDocument->readU8();
                if ( foo > 31 )
                    std::cerr << (char)foo;
                else if ( foo == PARAGRAPH_MARK )
                    std::cerr << std::endl;
                else if ( foo > 0 )
                    std::cerr << "{" <<  (int)foo << "}";
                else
                    std::cerr << "_";
            }
            std::cerr << std::endl << "   position: " << m_wordDocument->tell() << ", limit: " << limit << std::endl;
        }
#endif
    }
    else {
        std::cerr << "Oooops, couldn't find the piece table." << std::endl;
        return false;
    }
    return true;
}

bool Parser95::parseBody()
{
    m_subDocumentHandler->bodyStart();
    m_textHandler->sectionStart( SharedPtr<const Word97::SEP>( new Word97::SEP ) );  // looks reasonable, eh?

    // Read all the pieces belonging to the main document text
    Word95::PLCFIterator<Word95::PCD, 8> it( *m_plcfpcd );
    while ( m_currentCP < m_fib.ccpText && it.current() ) {
        U32 fc = it.current()->fc;
        U32 limit = it.currentRun() << 1;
        std::cerr << "   value: " << fc << std::endl;
        if ( fc & 0x40000000 ) {
            fc = ( fc & 0xbfffffff ) >> 1;
            limit >>= 1;
            std::cerr << "   value (cleared 2nd MSB, div. by 2): " << fc << std::endl;
        }
        m_wordDocument->seek(fc);
        std::cerr << "   position: " << m_wordDocument->tell() << ", limit: " << limit << std::endl;
        for ( unsigned int j = 0; j < limit; ++j ) {
            U8 foo = m_wordDocument->readU8();
            if ( foo > 31 )
                std::cerr << (char)foo;
            else if ( foo == PARAGRAPH_MARK )
                std::cerr << std::endl;
            else if ( foo > 0 )
                std::cerr << "{" <<  (int)foo << "}";
            else
                std::cerr << "_";
        }
        std::cerr << std::endl << "   position: " << m_wordDocument->tell() << ", limit: " << limit << std::endl;
        m_currentCP += it.currentRun();
        std::cerr << "currentCP: " << m_currentCP << ", ccpText: " << m_fib.ccpText << std::endl;
        ++it;
    }
    std::cerr << "currentCP: " << m_currentCP << ", ccpText: " << m_fib.ccpText << std::endl;

    // Implicit end of the section
    m_textHandler->sectionEnd();  // at least I think so
    m_subDocumentHandler->bodyEnd();
    return true;
}
