/* This file is part of the wvWare 2 project
   Copyright (C) 2001 S.R.Haque <srhaque@iee.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifndef WORD95_HELPER_H
#define WORD95_HELPER_H

#include <vector>
#include <iostream>
#include <global.h>
#include <ustring.h>
#include <word95_generated.h>

namespace wvWare {

class OLEStreamReader;
class OLEStreamWriter;

namespace Word95 {

class STTBF {

public:
    STTBF(U16 lid, OLEStreamReader *reader, bool preservePos=false);
    STTBF(U16 lid, const U8 *ptr);
    ~STTBF();

    unsigned int count() const { return m_strings.size(); }
    unsigned int extraDataLength() const { return m_extraDataLength; }

    UString firstString() const;
    UString nextString() const;
    UString prevString() const;
    UString lastString() const;
    UString stringAt(unsigned int index);

    const U8 *firstExtra() const;
    const U8 *nextExtra() const;
    const U8 *prevExtra() const;
    const U8 *lastExtra() const;
    const U8 *extraAt(unsigned int index) const;

private:
    // Internal helper methods to avoid code duplication in the CTORs
    void init(U16 lid, OLEStreamReader *reader, const U8 *ptr);
    U16 readU16(OLEStreamReader *reader, const U8 **ptr);
    U8 readU8(OLEStreamReader *reader, const U8 **ptr);

    std::vector<UString> m_strings;
    mutable std::vector<UString>::const_iterator m_stringIt;
    int m_extraDataLength;
    std::vector<U8*> m_extraData;
    mutable std::vector<U8*>::const_iterator m_extraIt;
};


template<class T, unsigned int sizeOfT> class PLF {

public:
    PLF(OLEStreamReader *reader, bool preservePos=false) {
        if(preservePos)
            reader->push();
        U32 count=reader->readU32();
        for(U32 i=0; i<count; ++i)
            m_items.push_back(new T(reader, false));
        if(preservePos)
            reader->pop();
        it=m_items.begin();
    }
    PLF(const U8 *ptr) {
        U32 count=readU32(ptr);
        ptr+=4;
        for(U32 i=0; i<count; ++i, ptr+=sizeOfT)
            m_items.push_back(new T(ptr));
        it=m_items.begin();
    }
    ~PLF() {
        for(it=m_items.begin(); it!=m_items.end(); ++it)
            delete *it;
    }

    unsigned int count() const { return m_items.size(); }

    const T *first() const { it=m_items.begin(); if(it!=m_items.end()) return *it;  return 0; }
    const T *next() const { if(it==m_items.end()) return 0; ++it; if(it!=m_items.end()) return *it; return 0; }
    const T *prev() const { if(m_items.size()==0) return 0; if(it!=m_items.begin()) --it; return *it; }
    const T *last() const { it=m_items.end(); if(it==m_items.begin()) return 0; --it; return *it; }
    const T *at(unsigned int index) const { if(index<m_items.size()) return m_items[index]; return 0; }

private:
    // don't copy or assign it
    PLF(const PLF<T, sizeOfT> &rhs);
    PLF<T, sizeOfT> &operator=(const PLF<T, sizeOfT> &rhs);

    std::vector<T*> m_items;
    mutable typename std::vector<T*>::const_iterator it;
};


template<class T, unsigned int sizeOfT> class PLCFIterator;

template<class T, unsigned int sizeOfT> class PLCF {

    friend class PLCFIterator<T, sizeOfT>;
public:
    PLCF(U32 length, OLEStreamReader *reader, bool preservePos=false) {
        if(preservePos)
            reader->push();
        U32 count=calculateCount(length);
        for(U32 i=0; i<count+1; ++i)  // n+1 CPs/FCs
            m_indices.push_back(reader->readU32());
        for(U32 i=0; i<count; ++i)  // n "T"s
            m_items.push_back(new T(reader, false));
        if(preservePos)
            reader->pop();
    }
    PLCF(U32 length, const U8 *ptr) {
        U32 count=calculateCount(length);
        for(U32 i=0; i<count+1; ++i, ptr+=4)  // n+1 CPs/FCs
            m_indices.push_back(readU32(ptr));
        for(U32 i=0; i<count; ++i, ptr+=sizeOfT)  // n "T"s
            m_items.push_back(new T(ptr));
    }
    ~PLCF() {
        typename std::vector<T*>::const_iterator it=m_items.begin();
        for( ; it!=m_items.end(); ++it)
            delete *it;
    }

private:
    // don't copy or assign it
    PLCF(const PLCF<T, sizeOfT> &rhs);
    PLCF<T, sizeOfT> &operator=(const PLCF<T, sizeOfT> &rhs);

    U32 calculateCount(U32 length) {
        if((length-4)%(sizeOfT+4)) {
            std::cerr << "Warning: PLCF size seems to be screwed" << std::endl;
            std::cerr << "Warning: length: " << length << ", size: " << sizeOfT << ", mod: " << (length-4)%(sizeOfT+4) << std::endl;
            return 0;
        }
        return (length-4)/(sizeOfT+4);
    }

    std::vector<U32> m_indices;
    std::vector<T*> m_items;
};


template<class T, unsigned int sizeOfT> class PLCFIterator {

public:
    PLCFIterator(const PLCF<T, sizeOfT> &plcf) : m_plcf(plcf) {
        m_itemIt=m_plcf.m_items.begin();
        m_indexIt=m_plcf.m_indices.begin();
    }

    unsigned int count() const { return m_plcf.m_items.count(); }
    bool isEmpty() const { return m_plcf.m_items.count()==0; }

    T *toFirst() {
        m_itemIt=m_plcf.m_items.begin();
        m_indexIt=m_plcf.m_indices.begin();
        if(m_itemIt!=m_plcf.m_items.end())
            return *m_itemIt;
        return 0;
    }

    // note: m_indexIt-=2 as we have n+1 indices!
    T *toLast() {
        m_itemIt=m_plcf.m_items.end();
        m_indexIt=m_plcf.m_indices.end();
        if(m_itemIt==m_plcf.m_items.begin())
            return 0;
        --m_itemIt;
        m_indexIt-=2;
        return *m_itemIt;
    }

    U32 currentStart() const { if(m_indexIt!=m_plcf.m_indices.end()) return *m_indexIt; return 0; }
    U32 currentLim() const { std::vector<U32>::const_iterator it=m_indexIt; if(it==m_plcf.m_indices.end()) return 0; ++it; return *it; }
    T *current() const { if(m_itemIt!=m_plcf.m_items.end()) return *m_itemIt; return 0; }

    U32 currentRun() const { return currentLim()-currentStart(); }

    T *operator++() {
        if(m_itemIt==m_plcf.m_items.end())
            return 0;
        ++m_itemIt;
        ++m_indexIt;
        if(m_itemIt!=m_plcf.m_items.end())
            return *m_itemIt;
        return 0;
    }
    T *operator--() {
        if(m_plcf.m_items.size()==0)
            return 0;
        if(m_itemIt!=m_plcf.m_items.begin()) {
            --m_itemIt;
            --m_indexIt;
        }
        return *m_itemIt;
    }

private:
    // don't copy or assign it
    PLCFIterator(const PLCFIterator<T, sizeOfT> &rhs);
    PLCFIterator<T, sizeOfT> &operator=(const PLCFIterator<T, sizeOfT> &rhs);

    const PLCF<T, sizeOfT> &m_plcf;
    typename std::vector<T*>::const_iterator m_itemIt;
    std::vector<U32>::const_iterator m_indexIt;
};


template<class T, class offset, unsigned int sizeOfOffset, offset (*readOffset) (OLEStreamReader *reader, const U8 *ptr)> class FKPIterator;

template<class T, class offset, unsigned int sizeOfOffset, offset (*readOffset) (OLEStreamReader *reader, const U8 *ptr)> class FKP {

    friend class FKPIterator<T, offset, sizeOfOffset, readOffset>;
public:
    FKP(OLEStreamReader *reader, bool preservePos=false) {
        if(preservePos)
            reader->push();
        reader->push();
        reader->seek(511, SEEK_CUR);
        m_crun=reader->readU8();
        reader->pop();

        m_rgfc=new U32[m_crun+1];
        for(U8 i=0; i<=m_crun; ++i)  // <= crun, because of crun+1 FCs!
            m_rgfc[i]=reader->readU32();

        m_rgb=new offset[m_crun];
        for(U8 i=0; i<m_crun; ++i)
            m_rgb[i]=readOffset(reader, 0);

        m_internalOffset=(m_crun+1)*4+m_crun*sizeOfOffset;

        // store the rest of the FKP in an internal array for later use
        U8 length=511-m_internalOffset;
        m_fkp=new U8 [length];  // 511, because we don't need crun
        for(U8 i=0; i<length; ++i)
            m_fkp[i]=reader->readU8();

        if(preservePos)
            reader->pop();
    }

    FKP(const U8 *ptr) {
        m_crun=ptr[511];

        m_rgfc=new U32[m_crun+1];
        for(U8 i=0; i<=m_crun; ++i, ptr+=4)  // <= crun, because of crun+1 FCs!
            m_rgfc[i]=readU32(ptr);

        m_rgb=new offset[m_crun];
        for(U8 i=0; i<m_crun; ++i, ptr+=sizeOfOffset)
            m_rgb[i]=readOffset(0, ptr);

        m_internalOffset=(m_crun+1)*4+m_crun*sizeOfOffset;

        // store the rest of the FKP in an internal array for later use
        U8 length=511-m_internalOffset;
        m_fkp=new U8 [length];  // 511, because we don't need crun
        for(U8 i=0; i<length; ++i, ++ptr)
            m_fkp[i]=*ptr;
    }

    ~FKP() {
        delete m_rgfc;
        delete m_rgb;
        delete m_fkp;
    }

    unsigned int crun() const { return m_crun; }
    bool isEmpty() const { return m_crun==0; }

private:
    // don't copy or assign it
    FKP(const FKP<T, offset, sizeOfOffset, readOffset> &rhs);
    FKP<T, offset, sizeOfOffset, readOffset> &operator=(const FKP<T, offset, sizeOfOffset, readOffset> &rhs);

    U8 m_crun;
    U32 *m_rgfc;  // array of FCs (crun+1)
    offset *m_rgb;  // array of offsets/BXs
    U16 m_internalOffset;  // offset to the start position of the "rest"
    U8 *m_fkp;  // the "rest" of the FKP
};

// Attention: This iterator has a non standard behavior of the current()
// method. Use it that way: for( ; !it.atEnd(); ++it)
template<class T, class offset, unsigned int sizeOfOffset, offset (*readOffset) (OLEStreamReader *reader, const U8 *ptr)> class FKPIterator {

public:
    FKPIterator(const FKP<T, offset, sizeOfOffset, readOffset> &fkp) : m_fkp(fkp), m_index(0) {}

    void toFirst() { m_index=0; }
    void toLast() { m_index=m_fkp.m_crun-1; }

    U32 currentStart() const { if(m_index<m_fkp.m_crun) return m_fkp.m_rgfc[m_index]; return 0; }
    U32 currentLim() const { if(m_index<m_fkp.m_crun) return m_fkp.m_rgfc[m_index+1]; return 0; }
    offset currentOffset() const  { if(m_index<m_fkp.m_crun) return m_fkp.m_rgb[m_index]; return offset(); }
    // Pointer to the start of the current CHPX/PAPX/..., 0 if we are at the end of the array
    // Attention: This iterator has a non standard behavior of the current()
    // method. Use it that way: for( ; !it.atEnd(); ++it)
    const U8 *current() const {
        if(m_index<m_fkp.m_crun) {
            // Note: The first byte of the "offset" types (BX or U8) is always
            // the word offset into the array (or 0!).
            const U8 *tmp=reinterpret_cast<const U8*>(&m_fkp.m_rgb[m_index]);
            // Now we have to calculate the real offset and then locate it
            // within our cached array...
            if(*tmp!=0)
                return &m_fkp.m_fkp[*tmp*2-m_fkp.m_internalOffset];
        }
        return 0;
    }

    void operator++() {
        if(m_index<m_fkp.m_crun)
            ++m_index;
    }

    void operator--() {
        if(m_index>0)
            --m_index;
    }

    U8 index() const { return m_index; }
    void setIndex(U8 index) { if(index<m_fkp.m_crun) m_index=index; }

    bool atEnd() { return m_index>=m_fkp.m_crun; }

private:
    // don't copy or assign it
    FKPIterator(const FKPIterator<T, offset, sizeOfOffset, readOffset> &rhs);
    FKPIterator<T, offset, sizeOfOffset, readOffset> &operator=(const FKPIterator<T, offset, sizeOfOffset, readOffset> &rhs);

    const FKP<T, offset, sizeOfOffset, readOffset> &m_fkp;
    U8 m_index;
};


/**
 * BX entry in a PAP FKP (1 byte word offset + 1 PHE)
 */
struct BX {
    /**
     * Creates an empty BX structure and sets the defaults
     */
    BX();
    /**
     * Simply calls read(...)
     */
    BX(OLEStreamReader *stream, bool preservePos=false);
    /**
     * Simply calls readPtr(...)
     */
    BX(const U8 *ptr);
    /**
     * Copy a BX
     */
    BX(const BX &rhs);

    BX &operator=(const BX &rhs);

    /**
     * This method reads the BX structure from the stream.
     * If  preservePos is true we push/pop the position of
     * the stream to save the state. If it's false the state
     * of stream will be changed!
     */
    bool read(OLEStreamReader *stream, bool preservePos=false);

    /**
     * This method reads the struct from a pointer
     */
    void readPtr(const U8 *ptr);

    /**
     * Same as reading :)
     */
    bool write(OLEStreamWriter *stream, bool preservePos=false);

    /**
     * Set all the fields to the inital value (default is 0)
     */
    void clear();

    // Data
    /**
     * One byte offset. This byte is a word offset, so that we can
     * address 512 bytes with one unsigned char and all the CHPX/PAPX
     * grpprls are aligned at 2 byte boundaries.
     */
    U8 offset;

    /**
     * The PHE for the paragraph described by this BX structure.
     */
    PHE phe;

}; // BX

bool operator==(const BX &lhs, const BX &rhs);
bool operator!=(const BX &lhs, const BX &rhs);

// Helper functions to read an U8 and a BX for the FKP template
namespace FKPWrapper {
U8 readFKPU8(OLEStreamReader *reader, const U8 *ptr);
BX readFKPBX(OLEStreamReader *reader, const U8 *ptr);
} // namespace FKPWrapper


// This enum is a "convenience enum" for reading the piece table
typedef enum {
    clxtGrpprl=1,
    clxtPlcfpcd=2
} clxtENUM;

} // namespace Word95
} // namespace wvWare

#endif // WORD95_HELPER_H
